#include <J3ML/Geometry/LineSegment.hpp>
#include <J3ML/Geometry/Sphere.hpp>
#include <J3ML/Geometry/TriangleMesh.hpp>

namespace J3ML::Geometry
{

    bool Sphere::Contains(const LineSegment &lineseg) const {
        return Contains(lineseg.A) && Contains(lineseg.B);
    }

    TriangleMesh Sphere::GenerateUVSphere() const
    {
        // TODO: Implement this later
        return TriangleMesh();
    }

    TriangleMesh Sphere::GenerateIcososphere() const
    {
        // TODO: Implement this later
        return TriangleMesh();
    }

    void Sphere::ProjectToAxis(const Vector3 &direction, float &outMin, float &outMax) const
    {
        float d = Vector3::Dot(direction, Position);
        outMin = d - Radius;
        outMax = d + Radius;
    }

    Vector3 Sphere::ExtremePoint(const Vector3 &direction) const {
        float len = direction.Length();
        assert(len > 0.f);
        return Position + direction * (Radius / len);
    }

    Vector3 Sphere::ExtremePoint(const Vector3 &direction, float &projectionDistance) const {
        Vector3 extremePoint = ExtremePoint(direction);
        projectionDistance = extremePoint.Dot(direction);
        return extremePoint;
    }

    void Sphere::Translate(const Vector3 &offset) {
        Position = Position + offset;
    }

    void Sphere::Transform(const Matrix3x3 &transform) {
        Position = transform * Position;
    }

    void Sphere::Transform(const Matrix4x4 &transform) {
        Position = transform * Position;
    }

    float Sphere::Cube(float inp) {
        return inp*inp*inp;
    }

    float Sphere::Volume() const {
        return 4.f * Math::Pi * Cube(Radius) / 3.f;
    }

    float Sphere::SurfaceArea() const {
        return 4.f * Math::Pi * Cube(Radius) / 3.f;
    }

    bool Sphere::IsFinite() const {
        return Position.IsFinite() && std::isfinite(Radius);
    }

    bool Sphere::IsDegenerate() const {
        return !(Radius > 0.f) || !Position.IsFinite();
    }

    bool Sphere::Contains(const Vector3 &point) const {
        return Position.DistanceSquared(point) <= Radius*Radius;
    }

    bool Sphere::Contains(const Vector3 &point, float epsilon) const {
        return Position.DistanceSquared(point) <= Radius*Radius + epsilon;
    }

    Vector3 Sphere::RandomPointOnSurface(RNG &rng) const {
        Vector3 v = Vector3::Zero;
        // Rejection sampling analysis: The unit sphere fills ~52.4% of the volume of its enclosing box,
        // so this loop is expected to take only very few iterations before succeeding.
        for (int i = 0; i < 1000; ++i)
        {
            v.x = rng.FloatNeg1_1();
            v.y = rng.FloatNeg1_1();
            v.z = rng.FloatNeg1_1();
            float lenSq = v.LengthSquared();
            if (lenSq >= 1e-6f && lenSq <= 1.f)
                return Position + (Radius / std::sqrt(lenSq)) * v;
        }
        // Astronomically small probability to reach here, and if we do so, the provided RNG must have been in a bad state.
        assert(false && "Sphere::RandomPointOnSurface failed!");

        // Failed to generate a point inside this sphere. Return an arbitrary point on the surface as fallback.
        return Position + Vector3(Radius, 0, 0);
    }

    Vector3 Sphere::RandomPointInside(RNG &rng) {
        assert(Radius > 1e-3f);
        Vector3 v = Vector3::Zero;

        // Rejection sampling analysis: The unit sphere fills ~52.4% of the volume of its enclosing box
        // so this loop is expected to take only very few iterations before succeeding.
        for (int i = 0; i < 1000; ++i)
        {
            v.x = rng.Float(-Radius, Radius);
            v.y = rng.Float(-Radius, Radius);
            v.z = rng.Float(-Radius, Radius);

            if (v.LengthSquared() <= Radius*Radius)
                return Position + v;
        }
        assert(false && "Sphere::RandomPointInside failed!");

        // Failed to generate a point inside this sphere. Return the sphere center as fallback.
        return Position;
    }


}
