/// Josh's 3D Math Library
/// A C++20 Library for 3D Math, Computer Graphics, and Scientific Computing.
/// Developed and Maintained by Josh O'Leary @ Redacted Software.
/// Special Thanks to William Tomasine II and Maxine Hayes.
/// (c) 2024 Redacted Software
/// This work is dedicated to the public domain.

/// @file AABBTests.hpp
/// @desc Unit tests for the Axis-Aligned Bounding Box class.
/// @edit 2024-07-06

#pragma once

#include <jtest/jtest.hpp>
#include <jtest/Unit.hpp>
#include <J3ML/Geometry/AABB.hpp>
#include <J3ML/Geometry/LineSegment.hpp>
#include <J3ML/Geometry/Sphere.hpp>

jtest::Unit AABBUnit {"AABB"};
namespace AABBTests {
    inline void Define()
    {
        using namespace jtest;
        using namespace J3ML::Geometry;

        AABBUnit += Test("Contains", [] {
            AABB a ({0,0,0}, {10,10,10});


            check(a.Contains({0,0,0}));
            check(a.Contains({10,10,10}));
            check(a.Contains({1,2,3}));
            check(!a.Contains({-1, 2, 3}));
            check(!a.Contains({1, -2, 3}));
            check(!a.Contains({1, 2, -3}));
            check(!a.Contains({11, 2, 3}));
        });

        AABBUnit += Test("ContainsAABB", [] {
            AABB a ({0,0,0}, {10,10,10});
            check(a.Contains(a));
            check(a.Contains(AABB({5, 5, 5}, {6, 6, 6})));
            check(a.Contains(AABB({5, 5, 5}, {10, 6, 6})));
            check(!a.Contains(AABB({5,5,5}, {15, 15, 15})));
            check(!a.Contains(AABB({5,5,5}, {5, 15, 5})));
            check(!a.Contains(AABB({-5,-5,-5}, {5, 5, 5})));
            check(!a.Contains(AABB({-5,-5,-5}, {0, 0, 0})));
        });


        AABBUnit += Test("ContainsLineSegment", [] {
            AABB a ({0,0,0}, {10,10,10});

            check(a.Contains(LineSegment({0,0,0}, {10,10,10})));
            check(a.Contains(LineSegment({10,10,10}, {0,0,0})));
            check(a.Contains(LineSegment({5,5,5}, {6,6,6})));
            check(a.Contains(LineSegment({5,5,5}, {10,6,6})));
            check(!a.Contains(LineSegment({5,5,5}, {15, 15, 15})));
            check(!a.Contains(LineSegment({5,5,5}, {5, 15,5})));
            check(!a.Contains(LineSegment({-5, -5, -5}, {5,5,5})));
            check(!a.Contains(LineSegment({-5, -5, -5}, {0,0,0})));
            check(!a.Contains(LineSegment({15, 15, 15}, {0,0,0})));
        });

        AABBUnit += Test("ContainsSphere", [] {
            AABB a ({0,0,0}, {10,10,10});
            check(a.Contains(Sphere({0,0,0}, 0.f)));
            check(a.Contains(Sphere({5,5,5}, 1.f)));
            check(!a.Contains(Sphere({5,5,5}, 15.f)));
            check(!a.Contains(Sphere({9,5,5}, 2.f)));
            check(!a.Contains(Sphere({1,5,5}, 2.f)));
            check(!a.Contains(Sphere({-10, -10, -10}, 1000.f)));
        });

        AABBUnit += Test("IntersectsAABB", [] {
            AABB a({0,0,0}, {10,10,10});
            AABB b({5,0,0}, {15,10,10});
            AABB c({-5,-5,-5}, {0,10,0});
            AABB d({20,20,20}, {30,30,30});
            AABB e({1,1,1}, {9,9,9});
            check(a.Intersects(a));
            check(a.Intersects(b));
            check(!a.Intersects(c));
            check(!a.Intersects(d));
            check(a.Intersects(e));
        });


        AABBUnit += Test("TransformAsAABB", []{ /* TODO: Implement Test Stub */ });
        AABBUnit += Test("IsDegenerate",    []{ /* TODO: Implement Test Stub */ });
        AABBUnit += Test("Volume",          []{ /* TODO: Implement Test Stub */ });
    }

    inline void Run() {
        AABBUnit.RunAll();
    }
}
